#include <MT4Orders.mqh> // https://www.mql5.com/ru/code/16006
#include <fxsaber\Virtual\Virtual.mqh> // https://www.mql5.com/ru/code/22577
#include <Symbol.mqh> // https://www.mql5.com/ru/code/18855

#include "Timer.mqh"

class TESTERBASE : public TIMER
{
private:
  const string TesterSymbol;

  int AmountTicksAdd;
  int Pos;

  long LastTime;

  MqlTick Ticks[];

  bool FlagVirtual;

  void NewTicks( void )
  {
    this.NewTicks(this.AmountTicksAdd);

    return;
  }

  datetime GetNextTime( const datetime TimeBegin, const int Offset ) const
  {
    const ENUM_TIMEFRAMES Periods[] = {PERIOD_CURRENT, PERIOD_M1, PERIOD_M5, PERIOD_M15, PERIOD_M30, PERIOD_H1, PERIOD_H4, PERIOD_D1};

    datetime Res = 0;

    for (int i = 0; !Res && (i < ::ArraySize(Periods)); i++)
    {
      const int Pos2 = ::iBarShift(this.TesterSymbol, Periods[i], TimeBegin);

      if (Pos2 != -1)
      {
        const int Offset2 = Offset / ::PeriodSeconds(Periods[i]);

        Res = ::iTime(this.TesterSymbol, Periods[i], (Pos2 > Offset2) ? Pos2 - Offset2 : 0);

        if (Res)
          Res += ::PeriodSeconds(Periods[i]);
      }
    }

    return(Res);
  }

  static long GetNextTimeVirtual()
  {
    MqlTick Tick;

    return(VIRTUAL::GetCurrentTick(Tick) ? (Tick.time_msc / 60000 + 1) * 60000 : 0); //  .
  }

protected:
  const datetime StartTime;
  const SYMBOL Symb;

  const STRATEGY Strategy;

  virtual void System( const MqlTick &NewTicks[] )
  {
    long NextTimeVirtual = TESTERBASE::GetNextTimeVirtual();
    const int Size = ::ArraySize(NewTicks);

    for (int i = 0; i < Size; i++)
      if (this.FlagVirtual)
      {
        this.FlagVirtual = !TESTERBASE::NextVirtual();

        VIRTUAL::NewTick(NewTicks[i], this.Strategy);
        NextTimeVirtual = TESTERBASE::GetNextTimeVirtual();
      }
      else
      {
        VIRTUAL::NewTick(NewTicks[i], this.Strategy);

        if ((NewTicks[i].time_msc >= NextTimeVirtual) && TESTERBASE::NextVirtual())
          NextTimeVirtual = TESTERBASE::GetNextTimeVirtual();
      }

    return;
  }

#define DAY (24 * 3600)
  virtual void NewTicks( const int Amount, const int Interval = DAY )
  {
    const int Size = ::ArraySize(this.Ticks);

    if (this.Pos >= Size)
    {
      if (Size)
        this.LastTime = this.Ticks[Size - 1].time_msc + 1;

      if (this.LastTime)
      {
        ::ArrayFree(this.Ticks);
        ::CopyTicksRange(this.TesterSymbol, this.Ticks, COPY_TICKS_INFO, this.LastTime, (ulong)this.GetNextTime(this.LastTime / 1000, Interval) * 1000);

        this.Pos = 0;
      }
    }

    if (this.Pos < ::ArraySize(this.Ticks))
    {
      MqlTick NewTicks[];

      this.Pos += ::ArrayCopy(NewTicks, this.Ticks, 0, this.Pos, Amount);

      this.Symb += NewTicks;
      this.System(NewTicks);
    }

    return;
  }
#undef DAY

  void NewTicks( const MqlTick &NewTicks[] )
  {
    this.Symb += NewTicks;

    this.System(NewTicks);

    return;
  }

  void OnTimerTester()
  {
    if (this.IsMyTime())
      this.NewTicks();

    return;
  }

  void PrevVirtual( void )
  {
    const int PosVirtual = VIRTUAL::Total() - 1;

    if (PosVirtual)
    {
      VIRTUAL::Delete();

      VIRTUAL::SelectByIndex(PosVirtual);
    }

    MqlTick Tick;
    this.FlagVirtual = VIRTUAL::GetCurrentTick(Tick);

    ::ArrayFree(this.Ticks);

    this.Pos = 0;
    this.LastTime = Tick.time_msc;

    if (this.Pos < ::CopyTicks(this.TesterSymbol, this.Ticks, COPY_TICKS_INFO, this.LastTime, 1))
    {
      MqlTick LastTick[1];

      if ((::CopyTicks(this.Symb.Name, LastTick, COPY_TICKS_ALL, 0, 1) == 1) && (LastTick[0].time_msc == this.Ticks[Pos].time_msc))
        this.Symb.DeleteTicks(LastTick[0].time_msc);

      this.Symb += this.Ticks[this.Pos++];
    }

    return;
  }

  static bool NextVirtual( const bool FlagCopy = true )
  {
    return((FlagCopy ? VIRTUAL::CopyTo(VIRTUAL::Create(-1)) : (VIRTUAL::Create() != -1)) && (VIRTUAL::SelectByIndex(VIRTUAL::Total())));
  }


public:
  TESTERBASE( const string SymbName, const datetime dStartTime, const string SymbOriginal, const STRATEGY sStrategy = NULL ) :
                                        StartTime(dStartTime),
                                        TesterSymbol(SymbOriginal),
                                        AmountTicksAdd(1),
                                        Pos(0),
                                        Symb(::StringSubstr(SymbName, 0, 30), NULL, _Symbol),
                                        Strategy(sStrategy),
                                        LastTime((long)dStartTime * 1000),
                                        FlagVirtual(true)
  {
  }

  void ChangeVelocity( const double Koef )
  {
    this.AmountTicksAdd = this.ChangeVelocity(this.AmountTicksAdd, Koef);

    return;
  }
};